#include <stdio.h>
#include <stdlib.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <utk/ubertk.h>
#include <vmath.h>
#include "cam.h"
#include "utkgl.h"


bool init_gui();

void disp(void);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void keyb_up(unsigned char key, int x, int y);
void skeyb(int key, int x, int y);
void skeyb_up(int key, int x, int y);
void mouse(int bn, int state, int x, int y);
void motion(int x, int y);
void passive(int x, int y);

bool show_phong = true;
bool show_lambert = true;

Vector3 ldir(0, 1, 0);
float shininess = 50.0;
int num_samples = 1000;

static utk::Window *win;
static utk::Container *utkroot;
static utk::TrackBall *tball;

int main(int argc, char **argv)
{
	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("lafortune sampling");

	glutDisplayFunc(disp);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutKeyboardUpFunc(keyb_up);
	glutSpecialFunc(skeyb);
	glutSpecialUpFunc(skeyb_up);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(passive);
	glutIdleFunc(glutPostRedisplay);

	glPointSize(2.0);
	glEnable(GL_POINT_SMOOTH);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_DEPTH_TEST);

	if(init_utkgl() == -1) {
		return 1;
	}
	init_gui();

	glutMainLoop();
	return 0;
}

bool init_gui()
{
	utk::gfx::color = utk_color;
	utk::gfx::clip = utk_clip;
	utk::gfx::image = utk_image;
	utk::gfx::rect = utk_rect;
	utk::gfx::line = utk_line;
	utk::gfx::text = utk_text;
	utk::gfx::text_spacing = utk_text_spacing;
	utk::gfx::text_width = utk_text_width;

	// create the windows
	utkroot = utk::init(800, 600);

	win = utk::create_window(utkroot, 550, 30, 240, 250, "lafortune phong");
	win->set_alpha(192);
	win->show();

	utk::Container *vbox = utk::create_vbox(win);

	utk::Container *hbox = utk::create_hbox(vbox);
	utk::create_label(hbox, "samples:");
	utk::Slider *slider = utk::create_slider(hbox, 0, 5000, &num_samples);
	slider->set_value(num_samples);
	slider->set_vis_decimal(1);

	utk::create_checkbox(vbox, "diffuse samples", show_lambert, &show_lambert);
	utk::create_checkbox(vbox, "specular samples", show_phong, &show_phong);

	hbox = utk::create_hbox(vbox);
	utk::create_label(hbox, "shininess:");
	utk::create_slider(hbox, 0, 100, &shininess)->set_value(shininess);

	utk::create_label(vbox, "light vector:");
	tball = new utk::TrackBall(64, 64);
	vbox->add_child(tball);

	return true;
}

Vector3 sample_lambert(const Vector3 &ldir)
{
	float phi = acos(sqrt(frand(1.0)));
	float theta = 2.0 * M_PI * frand(1.0);

	Vector3 v;
	v.x = cos(theta) * sin(phi);
	v.y = cos(phi);
	v.z = sin(theta) * sin(phi);

	float ndotl = dot_product(Vector3(0, 1, 0), ldir);

	return v * MAX(ndotl, 0.0);
}

Vector3 sample_phong(const Vector3 &ldir, float shininess)
{
	Matrix3x3 mat;
	Vector3 norm(0, 1, 0);
	Vector3 ref = ldir.reflection(norm);

	if(1.0 - dot_product(ldir, norm) > ERROR_MARGIN) {
		Vector3 ivec = cross_product(ldir, ref).normalized();
		Vector3 kvec = cross_product(ref, ivec);

		mat.set_column_vector(ivec, 0);
		mat.set_column_vector(ref, 1);
		mat.set_column_vector(kvec, 2);
	}

	float phi = acos(pow(frand(1.0), 1.0 / (shininess + 1)));
	float theta = 2.0 * M_PI * frand(1.0);

	Vector3 v;
	v.x = cos(theta) * sin(phi);
	v.y = cos(phi);
	v.z = sin(theta) * sin(phi);
	v.transform(mat);

	float vdotr = dot_product(ref, v);
	if(vdotr < 0.0) {
		vdotr = 0.0;
	}
	float mag = pow(vdotr, shininess);

	return v * mag;
}

void disp(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	set_view_matrix();

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glBegin(GL_QUADS);
	glColor3f(1, 1, 1);
	glVertex3f(-1, 0, -1);
	glVertex3f(1, 0, -1);
	glVertex3f(1, 0, 1);
	glVertex3f(-1, 0, 1);
	glEnd();
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glEnable(GL_BLEND);

	if(show_phong) {
		glBegin(GL_LINES);
		glColor3f(1, 0, 0);
		for(int i=0; i<num_samples; i++) {
			Vector3 dir = sample_phong(ldir, shininess);
			glVertex3f(0, 0, 0);
			glVertex3f(dir.x, dir.y, dir.z);
		}
		glEnd();
	}

	if(show_lambert) {
		glBegin(GL_POINTS);
		glColor3f(0, 1, 0);
		for(int i=0; i<num_samples; i++) {
			Vector3 dir = sample_lambert(ldir);
			glVertex3f(dir.x, dir.y, dir.z);
		}
		glEnd();
	}
	
	glDisable(GL_BLEND);

	float phi = tball->get_phi();
	float theta = tball->get_theta();

	ldir = SphVector(theta, phi);

	glBegin(GL_LINES);
	glColor3f(1, 1, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(ldir.x, ldir.y, ldir.z);
	glEnd();


	// draw the GUI
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);

	utk::draw(utkroot);

	glPopAttrib();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();


	glutSwapBuffers();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 1.0, 1000.0);
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	default:
		break;
	}

	utk::KeyboardEvent ev;
	ev.key = key;
	ev.pressed = true;

	event(&ev);
	glutPostRedisplay();
}

void keyb_up(unsigned char key, int x, int y)
{
	utk::KeyboardEvent ev;
	ev.key = key;
	ev.pressed = false;

	event(&ev);
	glutPostRedisplay();
}

static int glut_skey_to_keysym(int key)
{
	switch(key) {
	case GLUT_KEY_LEFT:
		return utk::KEY_LEFT;

	case GLUT_KEY_RIGHT:
		return utk::KEY_RIGHT;

	case GLUT_KEY_UP:
		return utk::KEY_UP;

	case GLUT_KEY_DOWN:
		return utk::KEY_DOWN;

	case GLUT_KEY_HOME:
		return utk::KEY_HOME;

	case GLUT_KEY_END:
		return utk::KEY_END;

	default:
		break;
	}

	return 0;
}

void skeyb(int key, int x, int y)
{
	if(!(key = glut_skey_to_keysym(key))) {
		return;
	}

	utk::KeyboardEvent ev;
	ev.key = key;
	ev.pressed = true;
	event(&ev);
	glutPostRedisplay();
}

void skeyb_up(int key, int x, int y)
{
	if(!(key = glut_skey_to_keysym(key))) {
		return;
	}

	utk::KeyboardEvent ev;
	ev.key = key;
	ev.pressed = false;
	event(&ev);
	glutPostRedisplay();
}


static int bnstate[16];
static int prev_x = -1, prev_y;

void mouse(int bn, int state, int x, int y)
{
	if(state == GLUT_UP) {
		bnstate[bn] = 0;
	}

	if(!win->hit_test(x, y)) {
		bnstate[bn] = state == GLUT_DOWN ? 1 : 0;

		if(state == GLUT_DOWN) {
			if(bn == 3) {
				cam_zoom(-10);
				glutPostRedisplay();
			} else if(bn == 4) {
				cam_zoom(10);
				glutPostRedisplay();
			} else {
				prev_x = x;
				prev_y = y;
			}
		} else {
			prev_x = -1;
		}
	}
	
	utk::MButtonEvent ev;
	ev.button = bn;
	ev.pressed = (state == GLUT_DOWN);
	ev.x = x;
	ev.y = y;

	event(&ev);
	glutPostRedisplay();
}

void motion(int x, int y)
{
	if(bnstate[0]) {
		cam_rotate(x - prev_x, y - prev_y);
		glutPostRedisplay();
	}
	if(bnstate[1]) {
		cam_pan(x - prev_x, y - prev_y);
		glutPostRedisplay();
	}
	if(bnstate[2]) {
		cam_zoom(y - prev_y);
		glutPostRedisplay();
	}

	prev_x = x;
	prev_y = y;

	utk::MMotionEvent ev;
	ev.x = x;
	ev.y = y;

	event(&ev);
	glutPostRedisplay();
}

void passive(int x, int y)
{
	motion(x, y);
}
